/* Name: main.c
 * Project: hid-data, example how to use HID for data transfer
 * Author: Christian Starkjohann
 * Creation Date: 2008-04-11
 * Tabsize: 4
 * Copyright: (c) 2008 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt), GNU GPL v3 or proprietary (CommercialLicense.txt)
 * This Revision: $Id: main.c 777 2010-01-15 18:34:48Z cs $
 *
 * Arranged for IME-display, by Eleken, 2012-04-11
 */

/*
This example should run on most AVRs with only little changes. No special
hardware resources except INT0 are used. You may have to change usbconfig.h for
different I/O pins for USB. Please note that USB D+ must be the INT0 pin, or
at least be connected to INT0 as well.
*/

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>  /* for sei(), timer0 */
#include <util/delay.h>     /* for _delay_ms() */
#include <avr/eeprom.h>

#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include "usbdrv.h"
#include "oddebug.h"        /* This is also an example for using debug macros */

#include "sreg.h"
#include "writer.h"

/* ------------------------------------------------------------------------- */

/* -- Extension -- */
volatile static uint8_t indicator_state;
volatile static uint8_t switch_state;
volatile static uint8_t switch_flags;


inline static void enableIntr(void){
//	TIMSK = 0b10000000; // enable OCIE2
}
inline static void disableIntr(void){
//	TIMSK = 0b00000000; // disable OCIE2
}

inline static void portInit(void){
	// ports
	DDRB  = 0b00001111;
	PORTB = 0b01000000;
	
	DDRC  = 0b00000000;
	PORTC = 0b00000000;
	
	DDRD  = (DDRD & 0b00001100)  | 0b11010011;
	PORTD = (PORTD & 0b00001100) | 0b01000010;
	
	// timer
	// timer2 (25 ms)
	TCCR2 = 0b00000111; // CTC, clk/1024
//	TIMSK = 0b10000000; // in enableIntr()
	OCR2 = 195;
	TCNT2 = 0;
}
/*
inline static void portUpdate(void){
	PORTB = (PORTB & 0b11000000) | (indicator_state & 0b00111111);
	
	if(switch_flags == 1){
		switch_state ^= 0b00000001;
		switch_flags = 0;
	}
}*/

// timer2 : 25ms
ISR(TIMER2_COMP_vect){
	static uint8_t sw_old;
//	static uint8_t s;
	static uint8_t cnt;
	
	if((cnt++ & 0x3) == 0){ // check switch state per 100ms
		uint8_t sw = (PINB & 0b10000000);
		
		if(sw_old == sw){ // switch state has not changed
//			sw_state = 0;
		}else { // (sw_old != sw) switch state has changed
			if((sw & 0b10000000) == 0){ // Switch A is pushed 
				switch_flags = 1;
			}else if(sw == 0b10000000){// Any switches are opened
//				switch_flags = s;
//				s = 0;
			}else {
//				s = 0;
			}
		}
		sw_old = sw;
	}
}


/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = {    /* USB report descriptor */
/*
	0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x40,                    //   REPORT_COUNT (64)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0xc0                           // END_COLLECTION
	*/
	
      0x06, 0xFF, 0xFF,         // 04|2   , Usage Page (vendordefined?)
      0x09, 0x01,               // 08|1   , Usage      (vendordefined
      0xA1, 0x01,               // A0|1   , Collection (Application)
      // IN report
      0x09, 0x02,               // 08|1   , Usage      (vendordefined)
      0x09, 0x03,               // 08|1   , Usage      (vendordefined)
      0x15, 0x00,               // 14|1   , Logical Minimum(0 for signed byte?)
      0x26 ,0xFF,0x00,           // 24|1   , Logical Maximum(255 for signed byte?)
      0x75, 0x08,               // 74|1   , Report Size(8) = field size in bits = 1 byte
      0x95, 0x40,			    // 94|1:ReportCount(size) = repeat count of previous item
      0x81, 0x02,               // 80|1: IN report (Data,Variable, Absolute)
      // OUT report
      0x09, 0x04,               // 08|1   , Usage      (vendordefined)
      0x09, 0x05,               // 08|1   , Usage      (vendordefined)
      0x15, 0x00,               // 14|1   , Logical Minimum(0 for signed byte?)
      0x26, 0xFF,0x00,           // 24|1   , Logical Maximum(255 for signed byte?)
      0x75, 0x08,               // 74|1   , Report Size(8) = field size in bits = 1 byte
      0x95, 0x40,			    // 94|1:ReportCount(size) = repeat count of previous item
      0x91, 0x02,               // 90|1: OUT report (Data,Variable, Absolute)
      // Feature report
      0x09, 0x06,               // 08|1   , Usage      (vendordefined)
      0x09, 0x07,               // 08|1   , Usage      (vendordefined)
      0x15, 0x00,               // 14|1   , LogicalMinimum(0 for signed byte)
      0x26, 0xFF,0x00,          // 24|1   , Logical Maximum(255 for signed byte)
      0x75, 0x08,               // 74|1   , Report Size(8) =field size in bits = 1 byte
      0x95, 0x08,               // 94|1:ReportCount
      0xB1, 0x02,               // B0|1:   Feature report
      0xC0                      // C0|0    , End Collection
      	
};
/* Since we define only one feature report, we don't use report-IDs (which
 * would be the first byte of the report). The entire report consists of 128
 * opaque data bytes.
 */

/* The following variables store the status of the current data transfer */
static uchar    currentPos;
static uchar    bytesRemaining;

volatile uint8_t g_rcvBuf[64];
volatile uint8_t g_sendBuf[64];

/* ------------------------------------------------------------------------- */

/* usbFunctionRead() is called when the host requests a chunk of data from
 * the device. For more information see the documentation in usbdrv/usbdrv.h.
 */
// Note: this function called every 8 bytes
// This function could be called even if the buffer is not updated
// Return 0xFF to abort with STALL
uchar   usbFunctionRead(uchar *data, uchar len)
{
	g_sendBuf[2] ++;
	
	if(len > bytesRemaining){
		len = bytesRemaining;
	}
	//    eeprom_read_block(data, (uchar *)0 + currentAddress, len);
	uint8_t i;
	for(i=0; i<len; i++){
		// test
		data[i] = g_sendBuf[currentPos];
		currentPos = (currentPos + 1) % 64;
	}
	//    currentAddress += len;
	bytesRemaining -= len;
	return len;
}

volatile uint8_t currentPos_rcv;
void usbInterruptRead(void)
{
	while(usbInterruptIsReady()){
		usbSetInterrupt((uint8_t*)&g_sendBuf[currentPos_rcv], 8);
		currentPos_rcv = (currentPos_rcv + 8) % 64;
	}
}

/* usbFunctionWrite() is called when the host sends a chunk of data to the
 * device. For more information see the documentation in usbdrv/usbdrv.h.
 */
// Note: this function called every 8 bytes
// Return 0xFF to abort with STALL
uchar   usbFunctionWrite(uchar *data, uchar len)
{
	if(bytesRemaining == 0)
		return 1;               /* end of transfer */
	if(len > bytesRemaining)
		len = bytesRemaining;
	//    eeprom_write_block(data, (uchar *)0 + currentAddress, len);
	uint8_t i;
	for(i=0; i<len; i++){
		g_rcvBuf[currentPos] = data[i]; // test
		currentPos = (currentPos + 1) % 64;
	}
	
	g_sendBuf[1] = len;
	
	bytesRemaining -= len;
	if(bytesRemaining == 0){ // This is the last chunk
		//usbDisableAllRequests();
		
		WriterFunc((const uint8_t*)g_rcvBuf, (uint8_t*) g_sendBuf); // This call should be moved outside of the core function
		//usbEnableAllRequests();
		
		return 1; /* return 1 if this was the last chunk */
	}
	
	return 0;
}

/* ------------------------------------------------------------------------- */

uint8_t idleRate;
usbMsgLen_t usbFunctionSetup(uchar data[8])
{
usbRequest_t    *rq = (void *)data;

    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* HID class request */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* since we have only one report type, we can ignore the report-ID */
            bytesRemaining = 64;
            currentPos = 0;
			//if((rq->wValue.word & 0xFF) == 1){
			//	usbMsgPtr = (uint8_t*)g_sendBuf;
			//	return 64;
			//}
            return USB_NO_MSG;  /* use usbFunctionRead() to obtain data */
        }else if(rq->bRequest == USBRQ_HID_SET_REPORT){
            /* since we have only one report type, we can ignore the report-ID */
            bytesRemaining = 64;
            currentPos = 0;
            return USB_NO_MSG;  /* use usbFunctionWrite() to receive data from host */
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /* ignore vendor type requests, we don't use any */
    	
    }
    return 0;
}

/* ------------------------------------------------------------------------- */

int main(void)
{
	uint8_t   i;

	wdt_enable(WDTO_1S);
	/* Even if you don't use the watchdog, turn it off here. On newer devices,
	 * the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
	 */
	/* RESET status: all port bits are inputs without pull-up.
	 * That's the way we need D+ and D-. Therefore we don't need any
	 * additional hardware initialization.
	 */
	odDebugInit();
	portInit();
	usbInit();
	usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
	i = 0;
	while(--i){             /* fake USB disconnect for > 250 ms */
		wdt_reset();
		_delay_ms(1);
	}
	usbDeviceConnect();
	sei();
	for(;;){                /* main event loop */
		wdt_reset();
		disableIntr();
		usbPoll();
		 // [Note] To use interrupt-in endpoint, poll for interrupt ready and set buffer in hand,,,
		usbInterruptRead();
		enableIntr();
		
		//portUpdate();
		//PORTB ^= 0b00001000; // Toggle PB3 (debug)
	}
    return 0;
}

/* ------------------------------------------------------------------------- */
